from .context import get_current_config, connections, log, run_task, metric
import time
import datetime
import gevent
import argparse
import random
import shlex
import traceback
from collections import defaultdict
from bson import ObjectId
try:
    from redis.lock import LuaLock
except ImportError:
    # Change name to avoid NameError raised when use of LuaLock at line 147
    from redis.lock import Lock as LuaLock
from .processes import Process, ProcessPool
from .utils import MovingETA, normalize_command
from .queue import Queue


class Agent(Process):
    """ MRQ Agent manages its local worker pool and takes turns in orchestrating the others in its group. """

    def __init__(self, worker_group=None):
        self.greenlets = {}
        self.id = ObjectId()
        self.worker_group = worker_group or get_current_config()["worker_group"]
        self.pool = ProcessPool(extra_env={
            "MRQ_AGENT_ID": str(self.id),
            "MRQ_WORKER_GROUP": self.worker_group
        })
        self.config = get_current_config()
        self.status = "started"
        metric("agent", data={"worker_group": self.worker_group, "agent_id": self.id})
        self.dateorchestrated = None

        # global redis key used to ensure only one agent orchestrator runs at a time
        self.redis_queuestats_lock_key = "%s:queuestatslock" % (self.config["redis_prefix"])

        # global HSET redis key used to store queue stats
        self.redis_queuestats_key= "%s:queuestats" % (self.config["redis_prefix"])

    def work(self):

        self.install_signal_handlers()
        self.datestarted = datetime.datetime.utcnow()

        self.pool.start()
        self.manage()

        self.greenlets["orchestrate"] = gevent.spawn(self.greenlet_orchestrate)
        self.greenlets["orchestrate"].start()

        self.greenlets["manage"] = gevent.spawn(self.greenlet_manage)
        self.greenlets["manage"].start()

        # Disabled for now
        # self.greenlets["queuestats"] = gevent.spawn(self.greenlet_queuestats)
        # self.greenlets["queuestats"].start()

        try:
            self.pool.wait()
        finally:
            self.shutdown_now()
            self.status = "stop"
            self.manage()

    def shutdown_now(self):
        self.pool.terminate()

        for g in self.greenlets.values():
            g.kill()

    def shutdown_graceful(self):
        self.pool.stop(timeout=None)

    def greenlet_manage(self):
        """ This greenlet always runs in background to update current status
            in MongoDB every N seconds.
         """

        while True:
            try:
                self.manage()
            except Exception as e:  # pylint: disable=broad-except
                log.error("When reporting: %s" % e)
            finally:
                time.sleep(self.config["report_interval"])

    def manage(self):

        report = self.get_agent_report()

        try:
            db = connections.mongodb_jobs.mrq_agents.find_and_modify({
                "_id": ObjectId(self.id)
            }, {"$set": report}, upsert=True)
            if not db:
                return
        except Exception as e:  # pylint: disable=broad-except
            log.debug("Agent report failed: %s" % e)
            return

        # If the desired_workers was changed by an orchestrator, apply the changes locally
        if self.status != "stop" and sorted(db.get("desired_workers", [])) != sorted(self.pool.desired_commands):

            group = self.fetch_worker_group_definition()
            process_termination_timeout = float(group.get("process_termination_timeout") or 60)
            self.pool.set_commands(db.get("desired_workers", []), timeout=process_termination_timeout)

    def get_agent_report(self):
        report = {
            "current_workers": [p["command"] for p in self.pool.processes],
            "total_cpu": get_current_config()["total_cpu"],
            "total_memory": get_current_config()["total_memory"],
            "worker_group": self.worker_group,
            "status": self.status,
            "dateorchestrated": self.dateorchestrated,
            "datestarted": self.datestarted,
            "datereported": datetime.datetime.utcnow(),
            "dateexpires": datetime.datetime.utcnow() + datetime.timedelta(seconds=(self.config["report_interval"] * 3) + 5)
        }
        metric("agent", data={"worker_group": self.worker_group, "agent_id": self.id, "worker_count": len(self.pool.processes)})
        return report

    def greenlet_orchestrate(self):
        while True:
            try:
                self.orchestrate()
            except Exception as e:
                log.error("Orchestration error! %s" % e)
                traceback.print_exc()

            time.sleep(self.config["orchestrate_interval"])

    def orchestrate(self):
        run_task("mrq.basetasks.orchestrator.Orchestrate", {})

    def greenlet_queuestats(self):

        interval = min(self.config["orchestrate_interval"], 1 * 60)
        lock_timeout = 5 * 60 + (interval * 2)

        while True:
            lock = LuaLock(connections.redis, self.redis_queuestats_lock_key,
                           timeout=lock_timeout, thread_local=False, blocking=False)
            with lock:
                lock_expires = time.time() + lock_timeout
                self.queue_etas = defaultdict(lambda: MovingETA(5))

                while True:
                    self.queuestats()

                    # Because queue stats can be expensive, we try to keep the lock on the same agent
                    lock_extend = (time.time() + lock_timeout) - lock_expires
                    lock_expires += lock_extend
                    lock.extend(lock_extend)

                    time.sleep(interval)

            time.sleep(interval)

    def queuestats(self):
        """ Compute ETAs for every known queue & subqueue """

        start_time = time.time()
        log.debug("Starting queue stats...")

        # Fetch all known queues
        queues = [Queue(q) for q in Queue.all_known()]

        new_queues = {queue.id for queue in queues}
        old_queues = set(self.queue_etas.keys())

        for deleted_queue in old_queues.difference(new_queues):
            self.queue_etas.pop(deleted_queue)

        t = time.time()
        stats = {}

        for queue in queues:
            cnt = queue.count_jobs_to_dequeue()
            eta = self.queue_etas[queue.id].next(cnt, t=t)

            # Number of jobs to dequeue, ETA, Time of stats
            stats[queue.id] = "%d %s %d" % (cnt, eta if eta is not None else "N", int(t))

        with connections.redis.pipeline(transaction=True) as pipe:
            if random.randint(0, 100) == 0 or len(stats) == 0:
                pipe.delete(self.redis_queuestats_key)
            if len(stats) > 0:
                pipe.hmset(self.redis_queuestats_key, stats)
            pipe.execute()

        log.debug("... done queue stats in %0.4fs" % (time.time() - start_time))

    def fetch_worker_group_definition(self):
        definition = connections.mongodb_jobs.mrq_workergroups.find_one({"_id": self.worker_group})

        # Prepend all commands by their worker profile.
        commands = []
        for command in definition.get("commands", []):
            simplified_command, worker_count = normalize_command(command, self.worker_group)
            commands.extend([simplified_command] * worker_count)

        definition["commands"] = commands
        return definition



from __future__ import print_function
from future.builtins import object
from future.utils import iteritems

from collections import defaultdict
import logging
import datetime
import sys
import pymongo
PY3 = sys.version_info > (3,)


def _encode_if_unicode(string):

    if PY3:
        return string

    if isinstance(string, unicode):  # pylint: disable=undefined-variable
        return string.encode("utf-8", "replace")
    else:
        return string


def _decode_if_str(string):

    if PY3:
        return str(string)

    if isinstance(string, str):
        return string.decode("utf-8", "replace")
    else:
        return unicode(string)  # pylint: disable=undefined-variable


class MongoHandler(logging.Handler):

    """ Job/Worker-aware log handler.

        We used the standard logging module before but it suffers from memory leaks
        when creating lots of logger objects.
    """

    def __init__(self, worker=None, mongodb_logs_size=16 * 1024 * 1024):
        super(MongoHandler, self).__init__()

        self.buffer = {}
        self.collection = None
        self.mongodb_logs_size = mongodb_logs_size

        self.reset()
        self.set_collection()
        # Import here to avoid import loop
        # pylint: disable=cyclic-import
        from .context import get_current_job, get_current_worker
        self.get_current_job = get_current_job
        self.worker = worker

    def set_collection(self):
        from .context import get_current_config, connections
        config = get_current_config()
        collection = config["mongodb_logs"]

        if collection == "1":
            self.collection = connections.mongodb_logs.mrq_logs
        if self.collection and self.mongodb_logs_size:
            if "mrq_logs" in connections.mongodb_logs.collection_names() and not self.collection.options().get("capped"):
                connections.mongodb_logs.command({"convertToCapped": "mrq_logs", "size": self.mongodb_logs_size})
            elif "mrq_logs" not in connections.mongodb_logs.collection_names():
                try:
                    connections.mongodb_logs.create_collection("mrq_logs", capped=True, size=self.mongodb_logs_size)
                except pymongo.errors.OperationFailure:  # The collection might have been created in the meantime
                    pass

    def reset(self):
        self.buffer = {
            "workers": defaultdict(list),
            "jobs": defaultdict(list)
        }

    def emit(self, record):
        log_entry = self.format(record)
        if self.collection is False:
            return
        log_entry = _decode_if_str(log_entry)

        if self.worker is not None:
            self.buffer["workers"][self.worker].append(log_entry)

        if record.name == "mrq.current":
            job_object = self.get_current_job()
            if job_object:
                self.buffer["jobs"][job_object.id].append(log_entry)

    def flush(self):
        # We may log some stuff before we are even connected to Mongo!
        if not self.collection:
            return

        inserts = [{
            "worker": k,
            "logs": "\n".join(v) + "\n"
        } for k, v in iteritems(self.buffer["workers"])] + [{
            "job": k,
            "logs": "\n".join(v) + "\n"
        } for k, v in iteritems(self.buffer["jobs"])]

        if len(inserts) == 0:
            return
        self.reset()

        try:
            self.collection.insert(inserts)
        except Exception as e:  # pylint: disable=broad-except
            self.emit("Log insert failed: %s" % e)



from future import standard_library
standard_library.install_aliases()
from future.builtins import next, map
from past.builtins import basestring
import logging
import gevent
import gevent.pool
import urllib.parse
import sys
import time
import pymongo
import traceback
from .utils import LazyObject, load_class_by_path
from .config import get_config
from .subpool import subpool_map, subpool_imap
from .exceptions import JobRuntimeError


# This should be MRQ's only Python object shared by all the jobs in the same process
_GLOBAL_CONTEXT = {

    # Contains all the running greenlets for this worker. greenletid => Job object
    "greenlets": {},

    # pointer to the current worker
    "worker": None,

    # pointer to the current config
    "config": {}
}

# Global log object, usable from all jobs
log = logging.getLogger("mrq.current")


def setup_context(**kwargs):
    """ Setup MRQ's environment.

        Note: gevent should probably be initialized too if you want to use concurrency.
    """
    set_current_config(get_config(**kwargs))


def set_current_job(job):
    current = gevent.getcurrent()

    current.__dict__["_trace_time"] = 0
    current.__dict__["_trace_switches"] = 0

    if job is None:
        if id(current) in _GLOBAL_CONTEXT["greenlets"]:
            del _GLOBAL_CONTEXT["greenlets"][id(current)]
    else:
        _GLOBAL_CONTEXT["greenlets"][id(current)] = (current, job)


def get_current_job(greenlet_id=None):
    if greenlet_id is None:
        greenlet_id = id(gevent.getcurrent())
    pair = _GLOBAL_CONTEXT["greenlets"].get(greenlet_id)
    if not pair:
        return None
    return pair[1]


def set_current_worker(worker):
    _GLOBAL_CONTEXT["worker"] = worker


def get_current_worker():
    return _GLOBAL_CONTEXT["worker"]

def set_logger_config():
    config = _GLOBAL_CONTEXT["config"]
    if config.get("quiet"):
        log.disabled = True
    else:
        log_format = config["log_format"]
        logging.basicConfig(format=log_format)
        log.setLevel(getattr(logging, config["log_level"]))

        handlers = config["log_handlers"].keys() if config["log_handlers"] else [config["log_handler"]]
        for handler in handlers:
            handler_class = load_class_by_path(handler)
            handler_config = config["log_handlers"].get(handler, {})
            handler_format = handler_config.pop("format", log_format)
            handler_level = getattr(logging, handler_config.pop("level", config["log_level"]))
            log_handler = handler_class(**handler_config)
            formatter = logging.Formatter(handler_format)
            log_handler.setFormatter(formatter)
            log_handler.setLevel(handler_level)
            log.addHandler(log_handler)


def set_current_config(config):
    _GLOBAL_CONTEXT["config"] = config

    if config["add_network_latency"] != "0" and config["add_network_latency"]:
        from mrq.monkey import patch_network_latency
        patch_network_latency(config["add_network_latency"])

    if config["print_mongodb"] or config["trace_io"]:
        from mrq.monkey import patch_pymongo
        patch_pymongo(config)

    if config["trace_io"]:
        from mrq.monkey import patch_io_all
        patch_io_all(config)


def get_current_config():
    if not _GLOBAL_CONTEXT["config"]:
        log.warning("get_current_config was called before setup of MRQ's environment. "
                    "Use context.setup_context() for setting up MRQ's environment.")
    return _GLOBAL_CONTEXT["config"]


def retry_current_job(delay=None, max_retries=None, queue=None, exception=None):
    current_job = get_current_job()
    if current_job:
        current_job.retry(delay=delay, max_retries=max_retries, queue=queue)
    elif exception:
        raise JobRuntimeError("Error during MRQ task execution") from exception
    else:
        raise JobRuntimeError("Error during MRQ task execution")


def abort_current_job():
    current_job = get_current_job()
    if current_job:
        current_job.abort()


def _connections_factory(attr):

    config = get_current_config()

    # Connection strings may be stored directly in config
    config_obj = config.get(attr)

    def versiontuple(v):
        return tuple(map(int, (v.split("."))))

    if attr.startswith("redis"):
        if isinstance(config_obj, basestring):

            import redis as pyredis
            import redis.connection as pyredisconnection

            connection_class = pyredisconnection.Connection
            urllib.parse.uses_netloc.append('redis')
            redis_url = urllib.parse.urlparse(config_obj)
            if redis_url.scheme == "rediss":
                connection_class = pyredisconnection.SSLConnection

            log.info("%s: Connecting to Redis at %s..." %
                     (attr, redis_url.hostname))

            redis_pool = pyredis.BlockingConnectionPool(
                host=redis_url.hostname,
                port=redis_url.port,
                db=int((redis_url.path or "").replace("/", "") or "0"),
                password=redis_url.password if redis_url.password is not None else redis_url.username,
                max_connections=int(config.get("redis_max_connections")),
                ssl_cert_reqs="none",
                timeout=int(config.get("redis_timeout")),
                decode_responses=False,
                connection_class=connection_class
            )
            return pyredis.StrictRedis(connection_pool=redis_pool)

        # Let's just assume we got a StrictRedis-like object!
        else:
            return config_obj

    elif attr.startswith("mongodb"):

        if isinstance(config_obj, basestring):

            if attr == "mongodb_logs" and config_obj == "1":
                return connections.mongodb_jobs
            elif config_obj == "0":
                return None

            from pymongo import MongoClient

            mongo_parsed = pymongo.uri_parser.parse_uri(config_obj)

            mongo_hosts = mongo_parsed["nodelist"]
            mongo_name = mongo_parsed["database"]

            log.debug("%s: Connecting to MongoDB at %s/%s..." % (attr, mongo_hosts, mongo_name))

            kwargs = {}

            db = MongoClient(config_obj, **kwargs)[mongo_name]

            log.debug("%s: ... connected. (readPreference=%s)" % (attr, db.read_preference))

            return db

        # Let's just assume we got a MongoDB-like object!
        else:
            return config_obj

connections = LazyObject()
connections.add_factory(_connections_factory)
del _connections_factory


def enable_greenlet_tracing():

    # Tracing seems to cause a 2-5% performance loss.

    import greenlet
    greenlet.GREENLET_USE_TRACING = True

    def trace(*args):

        time_since_last_switch = time.time() - trace.last_switch

        # Record the time of the current switch
        trace.last_switch = time.time()

        if args[0] == "switch":
            # We are switching from the greenlet args[1][0] to the greenlet
            # args[1][1]
            args[1][0].__dict__.setdefault("_trace_time", 0)
            args[1][0].__dict__["_trace_time"] += time_since_last_switch
            args[1][0].__dict__.setdefault("_trace_switches", 0)
            args[1][0].__dict__["_trace_switches"] += 1

        elif args[0] == "throw":
            pass

    trace.last_switch = time.time()

    greenlet.settrace(trace)  # pylint: disable=no-member


def run_task(path, params):
    """ Runs a task code synchronously """
    task_class = load_class_by_path(path)
    return task_class().run_wrapped(params)


def set_current_job_progress(ratio, save=False):
    job = get_current_job()
    if job:
        job.set_progress(ratio, save=save)


# Imports for backward compatibility
def queue_raw_jobs(*args, **kwargs):
    from . import job
    return job.queue_raw_jobs(*args, **kwargs)


def queue_job(*args, **kwargs):
    from . import job
    return job.queue_job(*args, **kwargs)


def queue_jobs(*args, **kwargs):
    from . import job
    return job.queue_jobs(*args, **kwargs)


def metric(*args, **kwargs):
    from . import helpers
    return helpers.metric(*args, **kwargs)



from __future__ import print_function
from future.utils import itervalues
from future.builtins import str
from mrq.task import Task
from mrq.queue import Queue
from bson import ObjectId
from mrq.context import connections, get_current_config, get_current_job
from mrq.job import set_queues_size
from collections import defaultdict
from mrq.utils import group_iter
import datetime
import ujson as json
import sys


def get_task_cfg(taskpath):
    return get_current_config().get("tasks", {}).get(taskpath) or {}


class JobAction(Task):

    params = None
    collection = None

    def run(self, params):

        self.params = params
        self.collection = connections.mongodb_jobs.mrq_jobs

        query = self.build_query()

        return self.perform_action(
            self.params.get("action"), query, self.params.get("destination_queue")
        )

    def build_query(self):
        query = {}
        current_job = get_current_job()

        if self.params.get("id"):
            query["_id"] = ObjectId(self.params.get("id"))

        # TODO use redis for queue
        for k in [
                "queue",
                "status",
                "worker",
                "path",
                "dateexpires",
                "dateretry",
                "exceptiontype"]:
            if self.params.get(k):
                if isinstance(self.params[k], (list, tuple)):
                    query[k] = {"$in": list(self.params[k])}
                else:
                    query[k] = self.params[k]
            if query.get("worker"):
                query["worker"] = ObjectId(query["worker"])

        if self.params.get("params"):
            params_dict = json.loads(self.params.get("params"))  # pylint: disable=no-member

            for key in params_dict:
                query["params.%s" % key] = params_dict[key]

        if current_job and "_id" not in query:
            query["_id"] = {"$lte": current_job.id}

        return query

    def perform_action(self, action, query, destination_queue):

        stats = {
            "requeued": 0,
            "cancelled": 0,
            "deleted": 0
        }

        if action == "cancel":

            default_job_timeout = get_current_config()["default_job_timeout"]

            # Finding the ttl here to expire is a bit hard because we may have mixed paths
            # and hence mixed ttls.
            # If we are cancelling by path, get this ttl
            if query.get("path"):
                result_ttl = get_task_cfg(query["path"]).get("result_ttl", default_job_timeout)

            # If not, get the maxmimum ttl of all tasks.
            else:

                tasks_defs = get_current_config().get("tasks", {})
                tasks_ttls = [cfg.get("result_ttl", 0) for cfg in itervalues(tasks_defs)]

                result_ttl = max([default_job_timeout] + tasks_ttls)

            now = datetime.datetime.utcnow()

            size_by_queues = defaultdict(int)
            if "queue" not in query:
                for job in self.collection.find(query, projection={"queue": 1}):
                    size_by_queues[job["queue"]] += 1

            ret = self.collection.update(query, {"$set": {
                "status": "cancel",
                "dateexpires": now + datetime.timedelta(seconds=result_ttl),
                "dateupdated": now
            }}, multi=True)
            stats["cancelled"] = ret["n"]

            if "queue" in query:
                if isinstance(query["queue"], str):
                    size_by_queues[query["queue"]] = ret["n"]
            set_queues_size(size_by_queues, action="decr")

            # Special case when emptying just by queue name: empty it directly!
            # In this case we could also loose some jobs that were queued after
            # the MongoDB update. They will be "lost" and requeued later like the other case
            # after the Redis BLPOP
            if list(query.keys()) == ["queue"]:
                Queue(query["queue"]).empty()

        elif action in ("requeue", "requeue_retry"):

            # Requeue task by groups of maximum 1k items (if all in the same
            # queue)
            cursor = self.collection.find(query, projection=["_id", "queue"])

            # We must freeze the list because queries below would change it.
            # This could not fit in memory, research adding {"stats": {"$ne":
            # "queued"}} in the query
            fetched_jobs = list(cursor)

            for jobs in group_iter(fetched_jobs, n=1000):

                jobs_by_queue = defaultdict(list)
                for job in jobs:
                    jobs_by_queue[job["queue"]].append(job["_id"])
                    stats["requeued"] += 1

                for queue in jobs_by_queue:
                    updates = {
                        "status": "queued",
                        "datequeued": datetime.datetime.utcnow(),
                        "dateupdated": datetime.datetime.utcnow()
                    }

                    if destination_queue is not None:
                        updates["queue"] = destination_queue

                    if action == "requeue":
                        updates["retry_count"] = 0

                    self.collection.update({
                        "_id": {"$in": jobs_by_queue[queue]}
                    }, {"$set": updates}, multi=True)
                if sys.version_info.major > 2:
                    set_queues_size({queue: len(jobs) for queue, jobs in jobs_by_queue.items()})
                else:
                    set_queues_size({queue: len(jobs) for queue, jobs in jobs_by_queue.iteritems()})
        elif action == 'delete':
            amount_delete = self.collection.delete_many(query)
            stats["deleted"] = amount_delete.deleted_count
        return stats




from setuptools import setup  # , find_packages
import os
import sys


def use_package(package):
    if not package:
        return False
    if package.startswith(('#', 'git+')):
        return False
    if sys.version_info.major > 2 and 'python_version <' in package:
        return False
    if sys.version_info.major == 2 and 'python_version >' in package:
        return False

    return True


def get_requirements():
    reqs = []
    for filename in ["requirements-base.txt", "requirements-dashboard.txt"]:
        with open(filename, "r") as f:
            reqs += [x.strip().split(";")[0] for x in f.readlines() if use_package(x.strip())]
    return reqs


def get_version():
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'mrq/version.py')) as f:
        locals = {}
        exec(f.read(), locals)
        return locals['VERSION']
    raise RuntimeError('No version info found.')

setup(
    name="mrq-custom",
    include_package_data=True,
    packages=['mrq', 'mrq.basetasks', 'mrq.bin', 'mrq.dashboard'],  # find_packages(exclude=['tests', 'tests.tasks']),
    version=get_version(),
    description="A simple yet powerful distributed worker task queue in Python",
    author="Pricing Assistant",
    license='MIT',
    author_email="contact@pricingassistant.com",
    url="http://github.com/pricingassistant/mrq",
    # download_url="http://chardet.feedparser.org/download/python3-chardet-1.0.1.tgz",
    keywords=["worker", "task", "distributed", "queue", "asynchronous", "redis", "mongodb", "job", "processing", "gevent"],
    platforms='any',
    entry_points={
        'console_scripts': [
            'mrq-worker = mrq.bin.mrq_worker:main',
            'mrq-run = mrq.bin.mrq_run:main',
            'mrq-agent = mrq.bin.mrq_agent:main',
            'mrq-dashboard = mrq.dashboard.app:main'
        ]
    },
    # dependency_links=[
    #     "http://github.com/mongodb/mongo-python-driver/archive/cb4adb2193a83413bc5545d89b7bbde4d6087761.zip#egg=pymongo-2.7rc1"
    # ],
    zip_safe=False,
    install_requires=get_requirements(),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        #'Development Status :: 1 - Planning',
        #'Development Status :: 2 - Pre-Alpha',
        #'Development Status :: 3 - Alpha',
        #'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        #'Development Status :: 6 - Mature',
        #'Development Status :: 7 - Inactive',
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities"
    ],
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
)



from future.builtins import str, object
from .context import log, queue_job
import datetime
import ujson as json
import time


def _hash_task(task):
    """ Returns a unique hash for identify a task and its params """

    params = task.get("params")
    if params:
        params = json.dumps(sorted(list(task["params"].items()), key=lambda x: x[0]))  # pylint: disable=no-member

    full = [str(task.get(x)) for x in ["path", "interval", "dailytime", "weekday", "monthday", "queue"]]

    full.extend([str(params)])
    return " ".join(full)


class Scheduler(object):

    def __init__(self, collection, config_tasks):
        self.collection = collection
        self.config_tasks = config_tasks
        self.config_synced = False
        self.all_tasks = []

    def check_config_integrity(self):
        """ Make sure the scheduler config is valid """
        tasks_by_hash = {_hash_task(t): t for t in self.config_tasks}

        if len(tasks_by_hash) != len(self.config_tasks):
            raise Exception("Fatal error: there was a hash duplicate in the scheduled tasks config.")

        for h, task in tasks_by_hash.items():
            if task.get("monthday") and not task.get("dailytime"):
                raise Exception("Fatal error: you can't schedule a task with 'monthday' and without 'dailytime' (%s)" % h)
            if task.get("weekday") and not task.get("dailytime"):
                raise Exception("Fatal error: you can't schedule a task with 'weekday' and without 'dailytime' (%s)" % h)

            if not task.get("monthday") and not task.get("weekday") and not task.get("dailytime") and not task.get("interval"):
                raise Exception("Fatal error: scheduler must be specified one of monthday,weekday,dailytime,interval. (%s)" % h)

    def sync_config_tasks(self):
        """ Performs the first sync of a list of tasks, often defined in the config file. """

        tasks_by_hash = {_hash_task(t): t for t in self.config_tasks}

        for task in self.all_tasks:
            if tasks_by_hash.get(task["hash"]):
                del tasks_by_hash[task["hash"]]
            else:
                self.collection.remove({"_id": task["_id"]})
                log.debug("Scheduler: deleted %s" % task["hash"])

        # What remains are the new ones to be inserted
        for h, task in tasks_by_hash.items():
            task["hash"] = h
            task["datelastqueued"] = datetime.datetime.fromtimestamp(0)
            if task.get("dailytime"):
                # Because MongoDB can store datetimes but not times,
                # we add today's date to the dailytime.
                # The date part will be discarded in check()
                task["dailytime"] = datetime.datetime.combine(
                    datetime.datetime.utcnow(), task["dailytime"])
                task["interval"] = 3600 * 24

                # Avoid to queue task in check() if today dailytime is already passed
                if datetime.datetime.utcnow().time() > task["dailytime"].time():
                    task["datelastqueued"] = datetime.datetime.utcnow()

            self.collection.find_one_and_update({"hash": task["hash"]}, {"$set": task}, upsert=True)
            log.debug("Scheduler: added %s" % task["hash"])

    def check(self):

        self.all_tasks = list(self.collection.find())

        if not self.config_synced:
            self.sync_config_tasks()
            self.all_tasks = list(self.collection.find())
            self.config_synced = True

        # log.debug(
        #     "Scheduler checking for out-of-date scheduled tasks (%s scheduled)..." %
        #     len(self.all_tasks)
        # )

        now = datetime.datetime.utcnow()
        current_weekday = now.weekday()
        current_monthday = now.day

        for task in self.all_tasks:

            interval = datetime.timedelta(seconds=task["interval"])

            if task["datelastqueued"] >= now:
                continue

            if task.get("monthday", current_monthday) != current_monthday:
                continue

            if task.get("weekday", current_weekday) != current_weekday:
                continue

            if task.get("dailytime"):
                if task["datelastqueued"].date() == now.date() or now.time() < task["dailytime"].time():
                    continue

            # if we only have "interval" key
            if all(k not in task for k in ["monthday", "weekday", "dailytime"]):
                if now - task["datelastqueued"] < interval:
                    continue

            queue_job(
                task["path"],
                task.get("params") or {},
                queue=task.get("queue")
            )

            self.collection.update({"_id": task["_id"]}, {"$set": {
                "datelastqueued": now
            }})

            log.debug("Scheduler: queued %s" % _hash_task(task))

        # Make sure we never again execute a scheduler with the same exact second.
        time.sleep(1)



from .queue import Queue
from . import context
import datetime
import copy
from pymongo.collection import ReturnDocument


class QueueRegular(Queue):

    def __init__(self, *args, **kwargs):
        Queue.__init__(self, *args, **kwargs)

        self.base_dequeue_query = {
            "status": "queued",
            "queue": self.id
        }

        whitelist = context.get_current_config().get("task_whitelist", "").strip()
        blacklist = context.get_current_config().get("task_blacklist", "").strip()

        if whitelist:
            self.base_dequeue_query["path"] = {"$in": [x.strip() for x in whitelist.split(",")]}
        elif blacklist:
            self.base_dequeue_query["path"] = {"$nin": [x.strip() for x in blacklist.split(",")]}

    @property
    def collection(self):
        return context.connections.mongodb_jobs.mrq_jobs

    def empty(self):
        """ Remove all jobs """
        return self.collection.delete_many({"queue": self.id})

    def get_retry_queue(self):
        """ Return the name of the queue where retried jobs will be queued """
        return self.id

    def get_known_subqueues(self):
        """ Returns all known subqueues """

        all_queues_from_mongodb = Queue.all_known(sources=("jobs", ))

        idprefix = self.id
        if not idprefix.endswith("/"):
            idprefix += "/"

        return {q for q in all_queues_from_mongodb if q.startswith(idprefix)}

    def size(self):
        """ Returns the total number of queued jobs on the queue """

        if self.id.endswith("/"):
            subqueues = self.get_known_subqueues()
            if len(subqueues) == 0:
                return 0
            else:
                with context.connections.redis.pipeline(transaction=False) as pipe:
                    for subqueue in subqueues:
                        pipe.get("queuesize:%s" % subqueue)
                    return [int(size or 0) for size in pipe.execute()]
        else:
            return int(context.connections.redis.get("queuesize:%s" % self.id) or 0)

    def list_job_ids(self, skip=0, limit=20):
        """ Returns a list of job ids on a queue """

        return [str(x["_id"]) for x in self.collection.find(
            {"status": "queued"},
            sort=[("_id", -1 if self.is_reverse else 1)],
            projection={"_id": 1})
        ]

    def dequeue_jobs(self, max_jobs=1, job_class=None, worker=None):
        """ Fetch a maximum of max_jobs from this queue """

        if job_class is None:
            from .job import Job
            job_class = Job

        count = 0



        # TODO: remove _id sort after full migration to datequeued
        sort_order = [("datequeued", -1 if self.is_reverse else 1), ("_id", -1 if self.is_reverse else 1)]

        # MongoDB optimization: with many jobs it's faster to fetch the IDs first and do the atomic update second
        # Some jobs may have been stolen by another worker in the meantime but it's a balance (should we over-fetch?)
        # job_ids = None
        # if max_jobs > 5:
        #     job_ids = [x["_id"] for x in self.collection.find(
        #         self.base_dequeue_query,
        #         limit=max_jobs,
        #         sort=sort_order,
        #         projection={"_id": 1}
        #     )]

        #     if len(job_ids) == 0:
        #         return

        for i in range(max_jobs):  # if job_ids is None else len(job_ids)):

            # if job_ids is not None:
            #     query = {
            #         "status": "queued",
            #         "_id": job_ids[i]
            #     }
            #     sort_order = None
            # else:
            query = self.base_dequeue_query

            job_data = self.collection.find_one_and_update(
                query,
                {"$set": {
                    "status": "started",
                    "datestarted": datetime.datetime.utcnow(),
                    "worker": worker.id if worker else None
                }, "$unset": {
                    "dateexpires": 1 # we don't want started jobs to expire unexpectedly
                }},
                sort=sort_order,
                return_document=ReturnDocument.AFTER,
                projection={
                    "_id": 1,
                    "path": 1,
                    "params": 1,
                    "status": 1,
                    "retry_count": 1,
                    "queue": 1,
                    "datequeued": 1
                }
            )

            if not job_data:
                break

            if worker:
                worker.status = "spawn"

            count += 1
            context.metric("queues.%s.dequeued" % job_data["queue"], 1)

            job = job_class(job_data["_id"], queue=self.id, start=False)
            job.set_data(job_data)
            job.datestarted = datetime.datetime.utcnow()

            context.metric("jobs.status.started")

            yield job

        context.metric("queues.all.dequeued", count)



from __future__ import division
from future.builtins import str, range, object
from past.utils import old_div
import re
import importlib
import time
import math
import json
import datetime
from collections import deque
from bson import ObjectId
import uuid
import shlex
#
# Utils are functions that should be independent from the rest of MRQ's codebase
#


def normalize_command(command, worker_group):
    if "--processes" in command:
        simplified_command = ""
        worker_count = 0
        skip_next = False
        for part in shlex.split(command):
            if skip_next:
                worker_count = part
                skip_next = False
                continue
            if part.startswith("--processes="):
                worker_count = part.split("=")[1]
                continue
            if part == "--processes":
                skip_next = True
                continue
            simplified_command += " %s" % part
            skip_next = False
        simplified_command = "MRQ_WORKER_GROUP=%s%s" % (worker_group, simplified_command)
        return simplified_command, int(worker_count)
    return "MRQ_WORKER_GROUP=%s %s" % (worker_group, command), 1


def get_local_ip():
    """ Returns the local IP. Can be overwritten in the config with --local-ip so don't call
    this function directly, instead get the current value from the config """
    import socket
    try:
        return socket.gethostbyname(socket.gethostname())
    except:  # pylint: disable=bare-except
        return "127.0.0.1"


def group_iter(iterator, n=2):
    """ Given an iterator, it returns sub-lists made of n items.
    (except the last that can have len < n)

    """

    # Use slices instead of an iterator when we have a flat list
    if isinstance(iterator, list):

        length = len(iterator)
        for i in range(int(math.ceil(old_div(float(length), n)))):
            yield iterator[i * n: (i + 1) * n]

    else:
        accumulator = []
        for item in iterator:
            accumulator.append(item)
            if len(accumulator) == n:
                yield accumulator
                accumulator = []

        # Yield what's left
        if len(accumulator) != 0:
            yield accumulator


# http://code.activestate.com/recipes/578231-probably-the-fastest-memoization-decorator-in-the-/
def memoize(f):
    """ Memoization decorator for a function taking one or more arguments. """
    class memodict(dict):

        def __getitem__(self, *key):
            return dict.__getitem__(self, key)

        def __missing__(self, key):
            ret = self[key] = f(*key)
            return ret

    return memodict().__getitem__


def memoize_single_argument(f):
    """ Memoization decorator for a function taking a single argument """
    class memodict(dict):

        def __missing__(self, key):
            ret = self[key] = f(key)
            return ret
    return memodict().__getitem__


@memoize_single_argument
def load_class_by_path(taskpath):
    """ Given a taskpath, returns the main task class. """

    return getattr(
        importlib.import_module(
            re.sub(
                r"\.[^.]+$",
                "",
                taskpath)),
        re.sub(
            r"^.*\.",
            "",
            taskpath))


def lazyproperty(fn):
    attr_name = '_lazy_' + fn.__name__

    @property
    def _lazyprop(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazyprop


# http://code.activestate.com/recipes/576655-wait-for-network-service-to-appear/
def wait_for_net_service(server, port, timeout=None, poll_interval=0.1):
    """ Wait for network service to appear
        @param timeout: in seconds, if None or 0 wait forever
        @return: True of False, if timeout is None may return only True or
                 throw unhandled network exception
    """
    import socket
    import errno

    s = socket.socket()
    if timeout:
        from time import time as now
        # time module is needed to calc timeout shared between two exceptions
        end = now() + timeout

    while True:
        try:
            if timeout:
                next_timeout = end - now()
                if next_timeout < 0:
                    return False
                else:
                    s.settimeout(next_timeout)

            s.connect((server, port))

        except socket.timeout as err:
            # this exception occurs only if timeout is set
            if timeout:
                return False

        except Exception as err:
            # catch timeout exception from underlying network library
            # this one is different from socket.timeout
            if not isinstance(err.args, tuple) or err.args[0] != errno.ETIMEDOUT:
                pass  # raise
        else:
            s.close()
            return True
        time.sleep(poll_interval)


class LazyObject(object):

    """ Lazy-connection class. Connections will only be initialized when first used. """

    def __init__(self):
        self._factories = []
        self._attributes_via_factories = []

    def add_factory(self, factory):
        self._factories.append(factory)

    # This will be called only once, when the attribute is still missing
    def __getattr__(self, attr):

        for factory in self._factories:
            value = factory(attr)
            if value is not None:
                self._attributes_via_factories.append(attr)
                self.__dict__[attr] = value
                return value

    def reset(self):
        # TODO proper connection close?
        for attr in self._attributes_via_factories:
            del self.__dict__[attr]
        self._attributes_via_factories = []


class MongoJSONEncoder(json.JSONEncoder):

    def default(self, obj):  # pylint: disable=E0202
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        elif isinstance(obj, ObjectId):
            return str(obj)
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, bytes):
            return obj.decode('utf-8')
        return json.JSONEncoder.default(self, obj)


class MovingAverage(object):

    def __init__(self, size):
        self.__size = size
        self.__sum = 0
        self.__q = deque([])

    def next(self, val):
        if len(self.__q) == self.__size:
            self.__sum -= self.__q.popleft()
        self.__sum += val
        self.__q.append(val)
        return 1.0 * self.__sum / len(self.__q)


class MovingETA(object):

    def __init__(self, size):
        self.__size = size
        self.__q = deque([])
        self.__t = deque([])

    def next(self, val, t=None):

        if t is None:
            t = time.time()

        if len(self.__q) == self.__size:
            self.__q.popleft()
            self.__t.popleft()
        self.__q.append(val)
        self.__t.append(t)

        if len(self.__q) == 1:
            return None

        mean_q = sum(self.__q) / len(self.__q)
        mean_t = sum(self.__t) / len(self.__t)

        def std(lst, m):
            return math.sqrt(sum((pow(x - m, 2) for x in lst)) / (len(lst) - 1))

        def pearson_r(list_t, list_q):

            sum_xy = 0
            sum_sq_v_x = 0
            sum_sq_v_y = 0

            for (x, y) in zip(list_t, list_q):
                var_x = x - mean_t
                var_y = y - mean_q
                sum_xy += var_x * var_y
                sum_sq_v_x += pow(var_x, 2)
                sum_sq_v_y += pow(var_y, 2)

            if sum_sq_v_x * sum_sq_v_y == 0:
                return None

            return sum_xy / math.sqrt(sum_sq_v_x * sum_sq_v_y)

        r = pearson_r(self.__t, self.__q)

        if r is None:
            return None

        # ax + b
        a = r * (std(self.__q, mean_q) / std(self.__t, mean_t))
        b = mean_q - a * mean_t

        # ETA is ax + b = 0
        if a == 0:
            return None

        return (-b / a) - t



from future.builtins import object

from .context import get_current_config, connections, log
from bson import ObjectId
from collections import defaultdict
from .processes import ProcessPool, Process


class Supervisor(Process):
    """ Manages several mrq-worker single processes """

    def __init__(self, command, numprocs=1):
        self.numprocs = numprocs
        self.command = command
        self.pool = ProcessPool()

    def work(self):

        self.install_signal_handlers()

        self.pool.set_commands([self.command] * self.numprocs)

        self.pool.start()

        self.pool.wait()

    def shutdown_now(self):
        self.pool.terminate()

    def shutdown_graceful(self):
        self.pool.stop(timeout=None)
